#!/usr/bin/env python3
# coding: utf-8

import os
import time
import backoff
import json
import requests
import pandas
import datetime
import logging
import healpy as hp
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from astropy.time import Time
from astropy import units as u
from astropy.coordinates import SkyCoord, Distance
from astropy.cosmology import Planck18 as cosmo
from ztfquery import alert, query, skyvision, io
from ztfquery import fields as ztfquery_fields
from matplotlib.backends.backend_pdf import PdfPages
from requests.auth import HTTPBasicAuth
from tqdm import tqdm
from ampel.ztf.t0.DecentFilter import DecentFilter
from ampel.ztf.dev.DevAlertProcessor import DevAlertProcessor
from ampel.alert.PhotoAlert import PhotoAlert
from ratelimit import limits, sleep_and_retry
from gwemopt.ztf_tiling import get_quadrant_ipix
from ampel.log.AmpelLogger import AmpelLogger
from nuztf.ampel_api import ampel_api_cone, ampel_api_name, reassemble_alert, ampel_api_catalog, ampel_api_tns

DEBUG = False
RATELIMIT_CALLS = 10
RATELIMIT_PERIOD = 1

class AmpelWizard:
    def __init__(
        self,
        run_config,
        t_min,
        logger=None,
        resource=None,
        filter_class=DecentFilter,
        cone_nside=64,
        cones_to_scan=None,
    ):
        self.cone_nside = cone_nside
        self.t_min = t_min

        if not hasattr(self, "prob_threshold"):
            self.prob_threshold = None

        if resource is None:
            resource = {
                "ampel-ztf/catalogmatch": "https://ampel.zeuthen.desy.de/api/catalogmatch/",
            }

        if logger is None:
            self.logger = AmpelLogger()
        else:
            self.logger = logger

        self.logger.info("AMPEL run config:")
        self.logger.info(run_config)

        self.ampel_filter_class = filter_class(
            logger=self.logger, resource=resource, **run_config
        )

        self.dap = DevAlertProcessor(self.ampel_filter_class)

        if not hasattr(self, "output_path"):
            self.output_path = None

        self.scanned_pixels = []
        if cones_to_scan is None:
            self.cone_ids, self.cone_coords = self.find_cone_coords()
        else:
            self.cone_ids, self.cone_coords = cones_to_scan
        self.cache = dict()
        self.default_t_max = t_min + 10.

        self.mns_time = str(self.t_min).split("T")[0].replace("-", "")
        self.mns = None

        self.overlap_prob = None
        self.overlap_fields = None
        self.first_obs = None
        self.last_obs = None
        self.n_fields = None
        self.area = None
        self.double_extragalactic_area = None

        if not hasattr(self, "dist"):
            self.dist = None

    def get_name(self):
        raise NotImplementedError

    def get_full_name(self):
        raise NotImplementedError

    @staticmethod
    def get_tiling_line():
        return ""

    @staticmethod
    def get_obs_line():
        raise NotImplementedError

    @staticmethod
    def remove_variability_line():
        raise NotImplementedError

    def get_overlap_line(self):
        raise NotImplementedError

    def filter_ampel(self, res):
        return (
            self.ampel_filter_class.apply(
                PhotoAlert(res["objectId"], res["objectId"], *self.dap._shape(res))
            )
            is not None
        )

    # @sleep_and_retry
    # @limits(calls=RATELIMIT_CALLS, period=RATELIMIT_PERIOD)
    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )

    def get_avro_by_name(self, ztf_name):
        return ampel_api_name(ztf_name, logger=self.logger)

    def add_to_cache_by_names(self, *args):
        for ztf_name in args:
            self.cache[ztf_name] = self.get_avro_by_name(ztf_name)

    def check_ampel_filter(self, ztf_name):
        lvl = logging.getLogger().getEffectiveLevel()
        logging.getLogger().setLevel(logging.DEBUG)
        self.logger.info("Set logger level to DEBUG")
        query_res = self.get_avro_by_name(ztf_name)
        self.logger.info("Checking filter f (no prv)")
        no_prv_bool = self.filter_f_no_prv(query_res)
        self.logger.info(f"Filter f (np prv): {no_prv_bool}")
        self.logger.info("Checking ampel filter")
        bool_ampel = self.filter_ampel(query_res)
        self.logger.info(f"ampel filter: {bool_ampel}")
        self.logger.info("Checking filter f (history)")
        history_bool = self.filter_f_history(query_res)
        self.logger.info(f"Filter f (history): {history_bool}")
        self.logger.info(f"Setting logger back to {lvl}")
        logging.getLogger().setLevel(lvl)
        return bool_ampel

    def plot_ztf_observations(self, **kwargs):
        self.get_multi_night_summary().show_gri_fields(**kwargs)

    def get_multi_night_summary(self, max_days=None):
        now = datetime.datetime.now()

        if max_days is not None:
            date_1 = datetime.datetime.strptime(self.mns_time, "%Y%m%d")
            end_date = date_1 + datetime.timedelta(days=max_days)
            end_date = end_date.strftime("%Y-%m-%d")
        else:
            end_date = now.strftime("%Y-%m-%d")

        if self.mns is None:
            start_date_jd = self.t_min.jd
            start_date_jd = Time(start_date_jd, format="jd").jd
            end_date_jd = Time(now).jd

            self.mns = skyvision.CompletedLog.from_daterange(
                self.mns_time, end=end_date, verbose=False
            )

            self.mns.data["obsjd"] = Time(
                list(self.mns.data.datetime.values), format="isot"
            ).jd

            self.mns.data.query(f"obsjd > {start_date_jd}", inplace=True)

            self.mns.data.reset_index(inplace=True)
            self.mns.data.drop(columns=["index"], inplace=True)

        return self.mns

    # def simulate_observations(self, fields):

    def scan_cones(self, t_max=None, max_cones=None):
        """ """

        if max_cones is None:
            max_cones = len(self.cone_ids)

        scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))

        self.logger.info("Commencing Ampel queries!")
        self.logger.info(f"Scan radius is {scan_radius}")
        self.logger.info(
            f"So far, {len(self.scanned_pixels)} pixels out of {len(self.cone_ids)} have already been scanned."
        )

        all_ztf_names = []

        for i, cone_id in enumerate(tqdm(list(self.cone_ids)[:max_cones])):
            ra, dec = self.cone_coords[i]

            if cone_id not in self.scanned_pixels:
                ztf_names = self.ampel_cone_search(
                    ra=np.degrees(ra),
                    dec=np.degrees(dec),
                    radius=scan_radius,
                    t_max=t_max,
                )

                if ztf_names:
                    for ztf_name in ztf_names:
                        all_ztf_names.append(ztf_name)

                self.scanned_pixels.append(cone_id)

        self.logger.info(f"Scanned {len(self.scanned_pixels)} pixels")

        # remove duplicates
        all_ztf_names = list(set(all_ztf_names))

        self.logger.info(f"Before filtering: Found {len(all_ztf_names)} candidates")
        self.logger.info(f"Retrieving alert history from AMPEL")

        results = self.ampel_object_search(
            ztf_names=all_ztf_names
        )

        for res in results:

            for res_alert in res:

                if res_alert["objectId"] not in self.cache.keys():
                    self.cache[res_alert["objectId"]] = res_alert
                elif (
                    res_alert["candidate"]["jd"]
                    > self.cache[res_alert["objectId"]]["candidate"]["jd"]
                ):
                    self.cache[res_alert["objectId"]] = res_alert

        self.logger.info(f"Found {len(self.cache)} candidates")

        self.create_candidate_summary()

    def filter_f_no_prv(self, res):
        raise NotImplementedError

    def fast_filter_f_no_prv(self, res):
        return self.filter_f_no_prv(res)

    def filter_f_history(self, res):
        raise NotImplementedError

    def fast_filter_f_history(self, res):
        return self.filter_f_history(res)

    def find_cone_coords(self):
        raise NotImplementedError

    @staticmethod
    def wrap_around_180(ra):
        ra[ra > np.pi] -= 2 * np.pi
        return ra

    # The backoff-stuff is to handle serverside
    # rate-limits when querying the API

    # @sleep_and_retry
    # @limits(calls=RATELIMIT_CALLS, period=RATELIMIT_PERIOD)
    @backoff.on_exception(
        backoff.expo,
        requests.exceptions.RequestException,
        max_time=600,
    )
    def ampel_cone_search(
        self, ra: float, dec: float, radius: float, t_max=None
    ) -> list:
        """ """
        if t_max is None:
            t_max = self.default_t_max

        t_min = self.t_min

        query_res = ampel_api_cone(ra, dec, radius, t_min.jd, t_max.jd, logger=self.logger)

        ## LEGACY (KEPT FOR TIMING RESULTS FOR JVS)
        # result = ampel_client.get_alerts_in_cone(
        #     ra=ra, dec=dec, radius=radius, jd_min=self.t_min.jd, jd_max=t_max.jd, with_history=False, max_blocks=100)
        # query_res = [i for i in result]

        ztf_names = []
        for res in query_res:
            if self.filter_f_no_prv(res):
                if self.filter_ampel(res):
                    ztf_names.append(res["objectId"])

        ztf_names = list(set(ztf_names))

        return ztf_names

    # @sleep_and_retry
    # @limits(calls=RATELIMIT_CALLS, period=RATELIMIT_PERIOD)

    def ampel_object_search(self, ztf_names: list) -> list:
        """ """
        all_results = []

        ## LEGACY (KEPT FOR TIMING RESULTS FOR JVS)
        # ztf_object = ampel_client.get_alerts_for_object(objectids, with_history=True)
        # query_res = [i for i in ztf_object]
        # query_res = self.merge_alerts(query_res)

        for ztf_name in ztf_names:

            query_res = ampel_api_name(ztf_name)

            final_res = []

            for res in query_res:
                if self.filter_f_history(res):
                    final_res.append(res)

            all_results.append(final_res)

        return all_results

    @staticmethod
    def calculate_abs_mag(mag, redshift: float):
        luminosity_distance = cosmo.luminosity_distance(redshift).value * 10 ** 6
        abs_mag = mag - 5 * (np.log10(luminosity_distance) - 1)
        return abs_mag

    def query_catalog(
        self,
        catalog: str,
        catalog_type: str,
        ra: float,
        dec: float,
        searchradius_arcsec: float = 10,
        searchtype: str = "all",
    ):
        """
        Method for querying catalogs via the Ampel API
        'catalog' must be the name of a supported catalog, e.g.
        SDSS_spec, PS1, NEDz_extcats...
        For a full list of catalogs, confer
        https://ampel.zeuthen.desy.de/api/catalogmatch/catalogs

        """
        assert catalog_type in ["extcats", "catsHTM"]
        assert searchtype in ["all", "nearest"]

        queryurl_catalogmatch = API_CATALOGMATCH_URL + "/cone_search/" + searchtype

        # First, we create a json body to post
        headers = {"accept": "application/json", "Content-Type": "application/json"}
        query = {
            "ra_deg": ra,
            "dec_deg": dec,
            "catalogs": [
                {"name": catalog, "rs_arcsec": searchradius_arcsec, "use": catalog_type}
            ],
        }

        response = requests.post(
            url=queryurl_catalogmatch, json=query, headers=headers
        ).json()[0]

        return response

    def query_ned_for_z(self, ra: float, dec: float, searchradius_arcsec: float = 20):

        z = None
        dist_arcsec = None

        query = ampel_api_catalog(
            catalog="NEDz_extcats",
            catalog_type="extcats",
            ra=ra,
            dec=dec,
            searchradius_arcsec=searchradius_arcsec,
            searchtype="nearest",
        )

        if query:
            z = query["body"]["z"]
            dist_arcsec = query["dist_arcsec"]
        return z, dist_arcsec

    def parse_candidates(self):

        table = (
            "+--------------------------------------------------------------------------------+\n"
            "| ZTF Name     | IAU Name  | RA (deg)    | DEC (deg)   | Filter | Mag   | MagErr |\n"
            "+--------------------------------------------------------------------------------+\n"
        )
        for name, res in sorted(self.cache.items()):

            jds = [x["jd"] for x in res["prv_candidates"]]

            if res["candidate"]["jd"] > max(jds):
                latest = res["candidate"]
            else:
                latest = res["prv_candidates"][jds.index(max(jds))]

            old_flag = ""

            second_det = [x for x in jds if x > min(jds) + 0.01]
            if len(second_det) > 0:
                if Time.now().jd - second_det[0] > 1.0:
                    old_flag = "(MORE THAN ONE DAY SINCE SECOND DETECTION)"

            tns_result = " ------- "
            tns_name, tns_date, tns_group = ampel_api_tns(
                latest["ra"], latest["dec"], searchradius_arcsec=3
            )
            if tns_name:
                tns_result = tns_name

            line = "| {0} | {1} | {2:011.7f} | {3:+011.7f} | {4}      | {5:.2f} | {6:.2f}   | {7} \n".format(
                name,
                tns_result,
                float(latest["ra"]),
                float(latest["dec"]),
                ["g", "r", "i"][latest["fid"] - 1],
                latest["magpsf"],
                latest["sigmapsf"],
                old_flag,
                sign="+",
                prec=7,
            )
            table += line

        table += "+--------------------------------------------------------------------------------+\n\n"
        return table

    def draft_gcn(self):

        text = (
            "Astronomer Name (Institute of Somewhere), ............. report,\n"
            "On behalf of the Zwicky Transient Facility (ZTF) and Global Relay of Observatories Watching Transients Happen (GROWTH) collaborations: \n"
            "We observed the localization region of the {0} with the Palomar 48-inch telescope, equipped with the 47 square degree ZTF camera (Bellm et al. 2019, Graham et al. 2019). {1}"
            "We started observations in the g-band and r-band beginning at {2} UTC, "
            "approximately {3:.1f} hours after event time. {4}"
            "{5} \n \n"
            "The images were processed in real-time through the ZTF reduction and image subtraction pipelines at IPAC to search for potential counterparts (Masci et al. 2019). "
            "AMPEL (Nordin et al. 2019, Stein et al. 2021) was used to search the alerts database for candidates. "
            "We reject stellar sources (Tachibana and Miller 2018) and moving objects, and "
            "apply machine learning algorithms (Mahabal et al. 2019) {6}. We are left with the following high-significance transient "
            "candidates by our pipeline, all lying within the "
            "{7}% localization of the skymap. \n\n{8} \n\n".format(
                self.get_full_name(),
                self.get_tiling_line(),
                self.first_obs.utc,
                (self.first_obs.jd - self.t_min.jd) * 24.0,
                self.get_overlap_line(),
                self.get_obs_line(),
                self.remove_variability_line(),
                100 * self.prob_threshold,
                self.parse_candidates(),
            )
        )

        if self.dist:
            text += (
                "The GW distance estimate is {:.0f} [{:.0f} - {:.0f}] Mpc.\n\n".format(
                    self.dist, self.dist - self.dist_unc, self.dist + self.dist_unc
                )
            )
        else:
            pass

        text += "Amongst our candidates, \n{0}. \n \n".format(self.text_summary())

        text += (
            "Based on observations obtained with the Samuel Oschin Telescope 48-inch and the 60-inch Telescope at the Palomar Observatory as part of the Zwicky Transient Facility project. ZTF is supported by the National Science Foundation under Grant No. AST-2034437 and a collaboration including Caltech, IPAC, the Weizmann Institute for Science, the Oskar Klein Center at Stockholm University, the University of Maryland, Deutsches Elektronen-Synchrotron and Humboldt University, the TANGO Consortium of Taiwan, the University of Wisconsin at Milwaukee, Trinity College Dublin, Lawrence Livermore National Laboratories, and IN2P3, France. Operations are conducted by COO, IPAC, and UW. \n"
            "GROWTH acknowledges generous support of the NSF under PIRE Grant No 1545949. \n"
            "Alert distribution service provided by DIRAC@UW (Patterson et al. 2019). \n"
            "Alert database searches are done by AMPEL (Nordin et al. 2019). \n"
            "Alert filtering is performed with the AMPEL Follow-up Pipeline (Stein et al. 2021). \n"
        )
        if self.dist:
            text += "Alert filtering and follow-up coordination is being undertaken by the Fritz marshal system (FIXME CITATION NEEDED)."
        return text

    @staticmethod
    def extract_ra_dec(nside, index):
        (colat, ra) = hp.pix2ang(nside, index, nest=True)
        dec = np.pi / 2.0 - colat
        return (ra, dec)

    @staticmethod
    def extract_npix(nside, ra, dec):
        colat = np.pi / 2.0 - dec
        return hp.ang2pix(nside, colat, ra, nest=True)

    def create_candidate_summary(self):

        self.logger.info(f"Saving to: {self.output_path}")

        with PdfPages(self.output_path) as pdf:
            for (name, old_alert) in tqdm(sorted(self.cache.items())):
                mock_alert = reassemble_alert(old_alert)
                try:
                    fig = alert.display_alert(mock_alert, show_ps_stamp=True)
                    fig.text(0.01, 0.01, name)
                    pdf.savefig()
                    plt.close()
                except TypeError:
                    self.logger.info(
                        f"WARNING!!! {name} will be missing from the report pdf for some reason."
                    )
                    pass

    @staticmethod
    def parse_ztf_filter(fid):
        return ["g", "r", "i"][fid - 1]

    def tns_summary(self):
        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]
            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]
            print(
                "Candidate:",
                name,
                res["candidate"]["ra"],
                res["candidate"]["dec"],
                first_detection["jd"],
            )
            try:
                last_upper_limit = [
                    x
                    for x in res["prv_candidates"]
                    if np.logical_and(
                        "isdiffpos" in x.keys(), x["jd"] < first_detection["jd"]
                    )
                ][-1]
                print(
                    "Last Upper Limit:",
                    last_upper_limit["jd"],
                    self.parse_ztf_filter(last_upper_limit["fid"]),
                    last_upper_limit["diffmaglim"],
                )
            except IndexError:
                last_upper_limit = None
                print("Last Upper Limit: None")
            print(
                "First Detection:",
                first_detection["jd"],
                self.parse_ztf_filter(first_detection["fid"]),
                first_detection["magpsf"],
                first_detection["sigmapsf"],
            )
            hours_after_merger = 24.0 * (first_detection["jd"] - self.t_min.jd)
            print(f"First observed {hours_after_merger} hours after merger")
            if last_upper_limit:
                print(
                    "It has risen",
                    -latest["magpsf"] + last_upper_limit["diffmaglim"],
                    self.parse_ztf_filter(latest["fid"]),
                    self.parse_ztf_filter(last_upper_limit["fid"]),
                )
            print(
                [
                    x["jd"]
                    for x in res["prv_candidates"] + [res["candidate"]]
                    if "isdiffpos" in x.keys()
                ]
            )
            print("\n")

    def candidate_text(self, name, first_detection, lul_lim, lul_jd):
        raise NotImplementedError

    def text_summary(self):
        text = ""
        for name, res in sorted(self.cache.items()):
            detections = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ]
            detection_jds = [x["jd"] for x in detections]
            first_detection = detections[detection_jds.index(min(detection_jds))]
            latest = [
                x
                for x in res["prv_candidates"] + [res["candidate"]]
                if "isdiffpos" in x.keys()
            ][-1]
            try:
                last_upper_limit = [
                    x
                    for x in res["prv_candidates"]
                    if np.logical_and(
                        "isdiffpos" in x.keys(), x["jd"] < first_detection["jd"]
                    )
                ][-1]

                text += self.candidate_text(
                    name,
                    first_detection["jd"],
                    last_upper_limit["diffmaglim"],
                    last_upper_limit["jd"],
                )

            # No pre-detection upper limit
            except IndexError:
                text += self.candidate_text(name, first_detection["jd"], None, None)

            ned_z, ned_dist = self.query_ned_for_z(
                latest["ra"], latest["dec"], searchradius_arcsec=20
            )

            if ned_z:
                ned_z = float(ned_z)
                absmag = self.calculate_abs_mag(latest["magpsf"], ned_z)
                if ned_z > 0:
                    z_dist = Distance(z=ned_z, cosmology=cosmo).value
                    text += f"It has a spec-z of {ned_z:.3f} [{z_dist:.0f} Mpc] and an abs. mag of {absmag:.1f}. Distance to SDSS galaxy is {ned_dist:.2f} arcsec. "
                    if self.dist:
                        gw_dist_interval = [
                            self.dist - self.dist_unc,
                            self.dist + self.dist_unc,
                        ]

            c = SkyCoord(res["candidate"]["ra"], res["candidate"]["dec"], unit="deg")
            g_lat = c.galactic.b.degree
            if abs(g_lat) < 15.0:
                text += f"It is located at a galactic latitude of {g_lat:.2f} degrees. "
            text += "\n"
        return text

    def simple_plot_overlap_with_observations(
        self, fields=None, first_det_window_days=None
    ):

        try:
            nside = self.ligo_nside
        except AttributeError:
            nside = self.nside

        fig = plt.figure()
        plt.subplot(projection="aitoff")

        probs = []
        single_probs = []

        if fields is None:
            mns = self.get_multi_night_summary(first_det_window_days)

        else:

            class MNS:
                def __init__(self, data):
                    self.data = pandas.DataFrame(
                        data, columns=["field", "ra", "dec", "datetime"]
                    )

            data = []

            for f in fields:
                ra, dec = ztfquery_fields.field_to_coords(f)[0]
                t = Time(Time.now().jd + 1.0, format="jd").utc
                t.format = "isot"
                t = t.value
                for _ in range(2):
                    data.append([f, ra, dec, t])

            mns = MNS(data)

        ras = np.degrees(
            self.wrap_around_180(
                np.array([np.radians(float(x)) for x in mns.data["ra"]])
            )
        )
        fields = list(mns.data["field"])

        plot_ras = []
        plot_decs = []

        single_ras = []
        single_decs = []

        veto_ras = []
        veto_decs = []

        overlapping_fields = []

        base_ztf_rad = 3.5
        ztf_dec_deg = 30.0

        for j, (ra, dec) in enumerate(tqdm(self.map_coords)):
            ra_deg = np.degrees(self.wrap_around_180(np.array([ra])))
            # ra_deg = self.wrap_around_180(np.array(np.degrees(ra)))
            dec_deg = np.degrees(dec)
            # (np.cos(dec - np.radians(ztf_dec_deg))
            ztf_rad = base_ztf_rad
            ztf_height = 3.7

            n_obs = 0

            for i, x in enumerate(mns.data["dec"]):
                if np.logical_and(
                    not dec_deg < float(x) - ztf_height,
                    not dec_deg > float(x) + ztf_height,
                ):
                    if abs(dec_deg - ztf_dec_deg) < 70.0:
                        if np.logical_and(
                            not ra_deg < float(ras[i]) - ztf_rad / abs(np.cos(dec)),
                            not ra_deg > float(ras[i]) + ztf_rad / abs(np.cos(dec)),
                        ):
                            n_obs += 1
                            fid = fields[i]
                            if fid not in overlapping_fields:
                                overlapping_fields.append(fields[i])

            if n_obs > 1:
                probs.append(self.map_probs[j])
                plot_ras.append(ra)
                plot_decs.append(dec)

            elif n_obs > 0:
                single_probs.append(self.map_probs[j])
                single_ras.append(ra)
                single_decs.append(dec)

            else:
                veto_ras.append(ra)
                veto_decs.append(dec)

        overlapping_fields = list(set(overlapping_fields))

        obs_times = np.array(
            [
                Time(mns.data["datetime"].iat[i], format="isot", scale="utc")
                for i in range(len(mns.data))
                if mns.data["field"].iat[i] in overlapping_fields
            ]
        )

        self.first_obs = min(obs_times)
        self.last_obs = max(obs_times)

        size = hp.max_pixrad(nside, degrees=True) ** 2

        # print(hp.max_pixrad(self.ligo_nside, degrees=True)**2 * np.pi, size)

        plt.scatter(
            self.wrap_around_180(np.array([plot_ras])),
            plot_decs,
            c=probs,
            vmin=0.0,
            vmax=max(self.data[self.key]),
            s=size,
        )

        plt.scatter(
            self.wrap_around_180(np.array([single_ras])),
            single_decs,
            c=single_probs,
            vmin=0.0,
            vmax=max(self.data[self.key]),
            s=size,
            cmap="gray",
        )

        plt.scatter(
            self.wrap_around_180(np.array([veto_ras])), veto_decs, color="red", s=size
        )

        red_patch = mpatches.Patch(color="red", label="Not observed")
        gray_patch = mpatches.Patch(color="gray", label="Observed once")
        plt.legend(handles=[red_patch, gray_patch])

        self.overlap_prob = 100.0 * np.sum(probs)
        once_observed_prob = 100 * (np.sum(probs) + np.sum(single_probs))
        message = (
            f"In total, {once_observed_prob} % of the contour was observed at least once. \n "
            f"In total, {self.overlap_prob} % of the contour was observed at least twice. \n"
            "THIS DOES NOT INCLUDE CHIP GAPS!!!"
        )

        print(message)

        self.area = (2.0 * base_ztf_rad) ** 2 * float(len(overlapping_fields))
        self.n_fields = len(overlapping_fields)
        self.overlap_fields = overlapping_fields

        print(
            f"{self.n_fields} fields were covered, covering approximately {self.area} sq deg."
        )
        return fig, message

    # def plot_overlap_with_observations(self, fields=None, pid=None, first_det_window_days=None, min_sep=0.01):
    #
    #     try:
    #         nside = self.ligo_nside
    #     except AttributeError:
    #         nside = self.nside
    #
    #     fig = plt.figure()
    #     plt.subplot(projection="aitoff")
    #
    #     double_in_plane_probs = []
    #     single_in_plane_prob = []
    #
    #     if fields is None:
    #         mns = self.get_multi_night_summary(first_det_window_days)
    #
    #     else:
    #
    #         class MNS:
    #             def __init__(self, data):
    #                 self.data = pandas.DataFrame(data, columns=["field", "ra", "dec", "datetime"])
    #
    #         data = []
    #
    #         for f in fields:
    #             ra, dec = ztfquery_fields.field_to_coords(f)[0]
    #             for i in range(2):
    #                 t = Time(self.t_min.jd + 0.1*i, format="jd").utc
    #                 t.format = "isot"
    #                 t = t.value
    #                 data.append([f, ra, dec, t])
    #
    #         mns = MNS(data)
    #
    #     data = mns.data.copy()
    #
    #     if pid is not None:
    #         pid_mask = data["pid"] == str(pid)
    #         data = data[pid_mask]
    #
    #     obs_times = np.array([Time(data["datetime"].iat[i], format="isot", scale="utc")
    #                           for i in range(len(data))])
    #
    #
    #     if first_det_window_days is not None:
    #         first_det_mask = [x < Time(self.t_min.jd + first_det_window_days, format="jd").utc for x in obs_times]
    #         data = data[first_det_mask]
    #         obs_times = obs_times[first_det_mask]
    #
    #
    #     pix_obs_times = dict()
    #
    #     print("Unpacking observations")
    #     pix_map = dict()
    #     for i, obs_time in enumerate(tqdm(obs_times)):
    #         radec = [f"{data['ra'].iat[i]} {data['dec'].iat[i]}"]
    #         coords = SkyCoord(radec, unit=(u.hourangle, u.deg))
    #         # pix = get_quadrant_ipix(nside, data["ra"].iat[i], data["dec"].iat[i])
    #         pix = get_quadrant_ipix(nside, coords.ra.deg[0], coords.dec.deg[0])
    #
    #         field = data["field"].iat[i]
    #
    #         flat_pix = []
    #
    #         for sub_list in pix:
    #             for p in sub_list:
    #                 flat_pix.append(p)
    #
    #         flat_pix = list(set(flat_pix))
    #
    #         t = obs_time.jd
    #         for p in flat_pix:
    #             if p not in pix_obs_times.keys():
    #                 pix_obs_times[p] = [t]
    #             else:
    #                 pix_obs_times[p] += [t]
    #
    #             if p not in pix_map.keys():
    #                 pix_map[p] = [field]
    #             else:
    #                 pix_map[p] += [field]
    #
    #
    #     npix = hp.nside2npix(nside)
    #     theta, phi = hp.pix2ang(nside, np.arange(npix), nest=False)
    #     radecs = SkyCoord(ra=phi * u.rad, dec=(0.5 * np.pi - theta) * u.rad)
    #     idx = np.where(np.abs(radecs.galactic.b.deg) <= 10.0)[0]
    #
    #     double_in_plane_pixels = []
    #     double_in_plane_probs = []
    #     single_in_plane_pixels = []
    #     single_in_plane_prob = []
    #     veto_pixels = []
    #     plane_pixels = []
    #     plane_probs = []
    #     times = []
    #     double_no_plane_prob = []
    #     double_no_plane_pixels = []
    #     single_no_plane_prob = []
    #     single_no_plane_pixels = []
    #
    #     overlapping_fields = []
    #     for i, p in enumerate(tqdm(hp.nest2ring(nside, self.pixel_nos))):
    #         if p in pix_obs_times.keys():
    #             if p in idx:
    #                 plane_pixels.append(p)
    #                 plane_probs.append(self.map_probs[i])
    #
    #             obs = pix_obs_times[p]
    #
    #             # check which healpix are observed twice
    #             if max(obs) - min(obs) > min_sep:
    #                 # is it in galactic plane or not?
    #                 if p not in idx:
    #                     double_no_plane_prob.append(self.map_probs[i])
    #                     double_no_plane_pixels.append(p)
    #                 else:
    #                     double_in_plane_probs.append(self.map_probs[i])
    #                     double_in_plane_pixels.append(p)
    #
    #             else:
    #                 if p not in idx:
    #                     single_no_plane_pixels.append(p)
    #                     single_no_plane_prob.append(self.map_probs[i])
    #                 else:
    #                     single_in_plane_prob.append(self.map_probs[i])
    #                     single_in_plane_pixels.append(p)
    #
    #             overlapping_fields += pix_map[p]
    #
    #             times += obs
    #         else:
    #             veto_pixels.append(p)
    #
    #     # print(f"double no plane prob = {double_no_plane_prob}")
    #     # print(f"probs = {probs}")
    #     # print(f"single no plane prob = {single_no_plane_prob}")
    #     # print(f"single probs = {single_probs}")
    #
    #     overlapping_fields = sorted(list(set(overlapping_fields)))
    #     self.overlap_fields = list(set(overlapping_fields))
    #
    #     self.overlap_prob = np.sum(double_in_plane_probs + double_no_plane_prob) * 100.
    #
    #     size = hp.max_pixrad(nside) ** 2 * 50.
    #
    #     veto_pos = np.array([hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in veto_pixels]).T
    #
    #     if len(veto_pos) > 0:
    #
    #         plt.scatter(self.wrap_around_180(np.radians(veto_pos[0])), np.radians(veto_pos[1]),
    #                     color="red", s=size)
    #
    #     plane_pos = np.array([hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in plane_pixels]).T
    #
    #     if len(plane_pos) > 0:
    #
    #         plt.scatter(self.wrap_around_180(np.radians(plane_pos[0])), np.radians(plane_pos[1]),
    #                     color="green", s=size)
    #
    #     single_pos = np.array([hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in single_no_plane_pixels]).T
    #
    #     if len(single_pos) > 0:
    #         plt.scatter(self.wrap_around_180(np.radians(single_pos[0])), np.radians(single_pos[1]),
    #                     c=single_no_plane_prob, vmin=0., vmax=max(self.data[self.key]), s=size, cmap='gray')
    #
    #     plot_pos = np.array([hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in double_no_plane_pixels]).T
    #
    #     if len(plot_pos) > 0:
    #         plt.scatter(self.wrap_around_180(np.radians(plot_pos[0])), np.radians(plot_pos[1]),
    #                     c=double_no_plane_prob, vmin=0., vmax=max(self.data[self.key]), s=size)
    #
    #     red_patch = mpatches.Patch(color='red', label='Not observed')
    #     gray_patch = mpatches.Patch(color='gray', label='Observed once')
    #     violet_patch = mpatches.Patch(color='green', label='Observed Galactic Plane (|b|<10)')
    #     plt.legend(handles=[red_patch, gray_patch, violet_patch])
    #
    #     message = "In total, {0:.2f} % of the contour was observed at least once. \n " \
    #               "This estimate includes {1:.2f} % of the contour " \
    #               "at a galactic latitude <10 deg. \n " \
    #               "In total, {2:.2f} % of the contour was observed at least twice. \n" \
    #               "In total, {3:.2f} % of the contour was observed at least twice, " \
    #               "and excluding low galactic latitudes. \n" \
    #               "These estimates accounts for chip gaps.".format(
    #         100 * (np.sum(double_in_plane_probs) + np.sum(single_in_plane_prob) + np.sum(single_no_plane_prob) + np.sum(double_no_plane_prob)),
    #         100 * np.sum(plane_probs),
    #         100.*(np.sum(double_in_plane_probs) + np.sum(double_no_plane_prob)),
    #         100.*np.sum(double_no_plane_prob)
    #         )
    #
    #     all_pix = single_in_plane_pixels + double_in_plane_pixels
    #
    #     n_pixels = len(single_in_plane_pixels + double_in_plane_pixels + double_no_plane_pixels + single_no_plane_pixels)
    #     n_double = len(double_no_plane_pixels + double_in_plane_pixels)
    #     n_plane = len(plane_pixels)
    #
    #     self.area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_pixels
    #     self.double_extragalactic_area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_double
    #     plane_area  = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_plane
    #     try:
    #         self.first_obs = Time(min(times), format="jd")
    #         self.first_obs.utc.format = "isot"
    #         self.last_obs = Time(max(times), format="jd")
    #         self.last_obs.utc.format = "isot"
    #
    #     except ValueError:
    #         raise Exception(f"No observations of this field were found at any time after {self.t_min.jd:.2f} JD ({self.t_min}). "
    #                         "Coverage overlap is 0%!")
    #
    #     print("Observations started at {0}".format(self.first_obs.jd))
    #
    #     self.overlap_fields = overlapping_fields
    #
    #     #     area = (2. * base_ztf_rad)**2 * float(len(overlapping_fields))
    #     #     n_fields = len(overlapping_fields)
    #
    #     print("{0} pixels were covered, covering approximately {1:.2g} sq deg.".format(
    #         n_pixels, self.area))
    #     print("{0} pixels were covered at least twice (b>10), covering approximately {1:.2g} sq deg.".format(
    #         n_double, self.double_extragalactic_area))
    #     print("{0} pixels were covered at low galactic latitude, covering approximately {1:.2g} sq deg.".format(
    #         n_plane, plane_area))
    #     return fig, message

    def plot_overlap_with_observations(
        self, fields=None, pid=None, first_det_window_days=None, min_sep=0.01
    ):

        try:
            nside = self.ligo_nside
        except AttributeError:
            nside = self.nside

        fig = plt.figure()
        plt.subplot(projection="aitoff")

        double_in_plane_probs = []
        single_in_plane_prob = []

        if fields is None:
            mns = self.get_multi_night_summary(first_det_window_days)

        else:

            class MNS:
                def __init__(self, data):
                    self.data = pandas.DataFrame(
                        data, columns=["field", "ra", "dec", "datetime"]
                    )

            data = []

            for f in fields:
                ra, dec = ztfquery_fields.get_field_centroid(f)[0]
                for i in range(2):
                    t = Time(self.t_min.jd + 0.1 * i, format="jd").utc
                    t.format = "isot"
                    t = t.value
                    data.append([f, ra, dec, t])

            mns = MNS(data)

        # Skip all 64 simultaneous quadrant entries, we only need one per observation for qa log
        # data = mns.data.copy().iloc[::64]

        data = mns.data.copy()

        ras = np.ones_like(data["field"]) * np.nan
        decs = np.ones_like(data["field"]) * np.nan

        # Actually load up ra/dec

        veto_fields = []

        for field in list(set(data["field"])):

            mask = data["field"] == field

            res = ztfquery_fields.get_field_centroid(field)

            if len(res) > 0:

                ras[mask] = res[0][0]
                decs[mask] = res[0][1]

            else:
                veto_fields.append(field)

        self.logger.info(
            f"No RA/Dec found by ztfquery for fields {veto_fields}. These observation have to be ignored."
        )

        data["ra"] = ras
        data["dec"] = decs

        mask = np.array([~np.isnan(x) for x in data["ra"]])

        data = data[mask]

        if pid is not None:
            pid_mask = data["pid"] == str(pid)
            data = data[pid_mask]

        obs_times = np.array(
            [
                Time(
                    data["datetime"].iat[i].replace(" ", "T"),
                    format="isot",
                    scale="utc",
                )
                for i in range(len(data))
            ]
        )

        if first_det_window_days is not None:
            first_det_mask = [
                x < Time(self.t_min.jd + first_det_window_days, format="jd").utc
                for x in obs_times
            ]
            data = data[first_det_mask]
            obs_times = obs_times[first_det_mask]

        pix_obs_times = dict()

        self.logger.info(f"Most recent observation found is {obs_times[-1]}")

        self.logger.info("Unpacking observations")
        pix_map = dict()

        for i, obs_time in enumerate(tqdm(obs_times)):

            # radec = [f"{data['ra'].iat[i]} {data['dec'].iat[i]}"]
            #
            # coords = SkyCoord(radec, unit=(u.deg, u.deg))
            pix = get_quadrant_ipix(nside, data["ra"].iat[i], data["dec"].iat[i])

            field = data["field"].iat[i]

            flat_pix = []

            for sub_list in pix:
                for p in sub_list:
                    flat_pix.append(p)

            flat_pix = list(set(flat_pix))

            t = obs_time.jd
            for p in flat_pix:
                if p not in pix_obs_times.keys():
                    pix_obs_times[p] = [t]
                else:
                    pix_obs_times[p] += [t]

                if p not in pix_map.keys():
                    pix_map[p] = [field]
                else:
                    pix_map[p] += [field]

        npix = hp.nside2npix(nside)
        theta, phi = hp.pix2ang(nside, np.arange(npix), nest=False)
        radecs = SkyCoord(ra=phi * u.rad, dec=(0.5 * np.pi - theta) * u.rad)
        idx = np.where(np.abs(radecs.galactic.b.deg) <= 10.0)[0]

        double_in_plane_pixels = []
        double_in_plane_probs = []
        single_in_plane_pixels = []
        single_in_plane_prob = []
        veto_pixels = []
        plane_pixels = []
        plane_probs = []
        times = []
        double_no_plane_prob = []
        double_no_plane_pixels = []
        single_no_plane_prob = []
        single_no_plane_pixels = []

        overlapping_fields = []
        for i, p in enumerate(tqdm(hp.nest2ring(nside, self.pixel_nos))):

            if p in pix_obs_times.keys():

                if p in idx:
                    plane_pixels.append(p)
                    plane_probs.append(self.map_probs[i])

                obs = pix_obs_times[p]

                # check which healpix are observed twice
                if max(obs) - min(obs) > min_sep:
                    # is it in galactic plane or not?
                    if p not in idx:
                        double_no_plane_prob.append(self.map_probs[i])
                        double_no_plane_pixels.append(p)
                    else:
                        double_in_plane_probs.append(self.map_probs[i])
                        double_in_plane_pixels.append(p)

                else:
                    if p not in idx:
                        single_no_plane_pixels.append(p)
                        single_no_plane_prob.append(self.map_probs[i])
                    else:
                        single_in_plane_prob.append(self.map_probs[i])
                        single_in_plane_pixels.append(p)

                overlapping_fields += pix_map[p]

                times += list(obs)
            else:
                veto_pixels.append(p)

        # print(f"double no plane prob = {double_no_plane_prob}")
        # print(f"probs = {probs}")
        # print(f"single no plane prob = {single_no_plane_prob}")
        # print(f"single probs = {single_probs}")

        overlapping_fields = sorted(list(set(overlapping_fields)))
        self.overlap_fields = list(set(overlapping_fields))

        self.overlap_prob = np.sum(double_in_plane_probs + double_no_plane_prob) * 100.0

        size = hp.max_pixrad(nside) ** 2 * 50.0

        veto_pos = np.array(
            [hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in veto_pixels]
        ).T

        if len(veto_pos) > 0:

            plt.scatter(
                self.wrap_around_180(np.radians(veto_pos[0])),
                np.radians(veto_pos[1]),
                color="red",
                s=size,
            )

        plane_pos = np.array(
            [hp.pixelfunc.pix2ang(nside, i, lonlat=True) for i in plane_pixels]
        ).T

        if len(plane_pos) > 0:

            plt.scatter(
                self.wrap_around_180(np.radians(plane_pos[0])),
                np.radians(plane_pos[1]),
                color="green",
                s=size,
            )

        single_pos = np.array(
            [
                hp.pixelfunc.pix2ang(nside, i, lonlat=True)
                for i in single_no_plane_pixels
            ]
        ).T

        if len(single_pos) > 0:
            plt.scatter(
                self.wrap_around_180(np.radians(single_pos[0])),
                np.radians(single_pos[1]),
                c=single_no_plane_prob,
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
                cmap="gray",
            )

        plot_pos = np.array(
            [
                hp.pixelfunc.pix2ang(nside, i, lonlat=True)
                for i in double_no_plane_pixels
            ]
        ).T

        if len(plot_pos) > 0:
            plt.scatter(
                self.wrap_around_180(np.radians(plot_pos[0])),
                np.radians(plot_pos[1]),
                c=double_no_plane_prob,
                vmin=0.0,
                vmax=max(self.data[self.key]),
                s=size,
            )

        red_patch = mpatches.Patch(color="red", label="Not observed")
        gray_patch = mpatches.Patch(color="gray", label="Observed once")
        violet_patch = mpatches.Patch(
            color="green", label="Observed Galactic Plane (|b|<10)"
        )
        plt.legend(handles=[red_patch, gray_patch, violet_patch])

        message = (
            "In total, {0:.2f} % of the contour was observed at least once. \n "
            "This estimate includes {1:.2f} % of the contour "
            "at a galactic latitude <10 deg. \n "
            "In total, {2:.2f} % of the contour was observed at least twice. \n"
            "In total, {3:.2f} % of the contour was observed at least twice, "
            "and excluding low galactic latitudes. \n"
            "These estimates accounts for chip gaps.".format(
                100
                * (
                    np.sum(double_in_plane_probs)
                    + np.sum(single_in_plane_prob)
                    + np.sum(single_no_plane_prob)
                    + np.sum(double_no_plane_prob)
                ),
                100 * np.sum(plane_probs),
                100.0 * (np.sum(double_in_plane_probs) + np.sum(double_no_plane_prob)),
                100.0 * np.sum(double_no_plane_prob),
            )
        )

        all_pix = single_in_plane_pixels + double_in_plane_pixels

        n_pixels = len(
            single_in_plane_pixels
            + double_in_plane_pixels
            + double_no_plane_pixels
            + single_no_plane_pixels
        )
        n_double = len(double_no_plane_pixels + double_in_plane_pixels)
        n_plane = len(plane_pixels)

        self.area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_pixels
        self.double_extragalactic_area = (
            hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_double
        )
        plane_area = hp.pixelfunc.nside2pixarea(nside, degrees=True) * n_plane

        try:
            self.first_obs = Time(min(times), format="jd")
            self.first_obs.utc.format = "isot"
            self.last_obs = Time(max(times), format="jd")
            self.last_obs.utc.format = "isot"

        except ValueError:
            raise Exception(
                f"No observations of this field were found at any time between {self.t_min} and"
                f"{obs_times[-1]}. Coverage overlap is 0%, but recent observations might be missing!"
            )

        self.logger.info(f"Observations started at {self.first_obs.jd}")

        self.overlap_fields = overlapping_fields

        #     area = (2. * base_ztf_rad)**2 * float(len(overlapping_fields))
        #     n_fields = len(overlapping_fields)

        print(
            f"{n_pixels} pixels were covered, covering approximately {self.area:.2g} sq deg."
        )
        print(
            f"{n_double} pixels were covered at least twice (b>10), covering approximately {self.double_extragalactic_area:.2g} sq deg."
        )
        print(
            f"{n_plane} pixels were covered at low galactic latitude, covering approximately {plane_area:.2g} sq deg."
        )
        return fig, message

    def crosscheck_prob(self):

        try:
            nside = self.ligo_nside
        except AttributeError:
            nside = self.nside

        class MNS:
            def __init__(self, data):
                self.data = pandas.DataFrame(
                    data, columns=["field", "ra", "dec", "datetime"]
                )

        data = []

        for f in self.overlap_fields:
            ra, dec = ztfquery_fields.field_to_coords(float(f))[0]
            t = Time(self.t_min.jd, format="jd").utc
            t.format = "isot"
            t = t.value
            data.append([f, ra, dec, t])

            mns = MNS(data)

        data = mns.data.copy()

        self.logger.info("Unpacking observations")
        field_prob = 0.0

        ps = []

        for index, row in tqdm(data.iterrows()):
            pix = get_quadrant_ipix(nside, row["ra"], row["dec"])

            flat_pix = []

            for sub_list in pix:
                for p in sub_list:
                    flat_pix.append(p)

            flat_pix = list(set(flat_pix))
            ps += flat_pix

        ps = list(set(ps))

        for p in hp.ring2nest(nside, ps):
            field_prob += self.data[self.key][int(p)]

        self.logger.info(
            f"Intergrating all fields overlapping 90% contour gives {100*field_prob:.2g}%"
        )

    def export_fields(self):
        mask = np.array([x in self.overlap_fields for x in self.mns.data["field"]])
        lim_mag = [20.5 for _ in range(np.sum(mask))]
        coincident_obs = self.mns.data[mask].assign(lim_mag=lim_mag)
        print(
            coincident_obs[["field", "pid", "datetime", "lim_mag", "exp"]].to_csv(
                index=False, sep=" "
            )
        )



import numpy as np
import os
from slack import RTMClient
from slack_bot import bot_access_token, access_token

submit_file = os.path.join(os.path.dirname(os.path.abspath(__file__)) + "/spawn_tmux_session.sh")

ampel_bot_user = ["UMNJK00CU", "DMBKJG00K"]

keywords = ["<@{0}>".format(ampel_bot_user), "LIGO", "banana", "GRB", "Fermi", "Ampel"]

def run_on_event(data):
    print(data)
    ts = data['ts']
    channel = data['channel']
    sid = int(float(ts) * 1.e6)
    cmd = "bash {0} {1} {2} {3}".format(submit_file, sid, ts, channel)
    print(cmd)
    os.system(cmd)

@RTMClient.run_on(event="message")
def say_hello(**payload):
    data = payload['data']
    if "user" in data.keys():
        try:
            if not np.logical_and(np.sum([x in data['text'] for x in keywords]) == 0, "DMBKJG00K" not in data["user"]):
                run_on_event(data)
        except KeyError:
            pass

if __name__ == "__main__":
    print("Running master client!")
    rtm_client = RTMClient(token=bot_access_token)
    rtm_client.start()



from nuztf.ampel_magic import AmpelWizard
from astropy.time import Time
import matplotlib.pyplot as plt
import healpy as hp
import numpy as np
from tqdm import tqdm
from ligo.gracedb.rest import GraceDb
import os
import requests
import lxml.etree
from astropy_healpix import HEALPix
from astropy.coordinates import SkyCoord
import fitsio
from astropy import units as u
import wget
from pathlib import Path
from numpy.lib.recfunctions import append_fields

# Setup LIGO client

ligo_client = GraceDb()

try:
    r = ligo_client.ping()
except HTTPError as e:
    raise(e.message)

base_ligo_dir = os.path.join(Path(__file__).resolve().parents[1], "../LIGO_skymaps")
ligo_candidate_output_dir = os.path.join(Path(__file__).resolve().parents[1], "../LIGO_candidates")

gw_run_config = {
    "min_ndet": 1,  # Default:2
    "min_tspan": -1,  # Default 0, but that rejects everything!
    "max_tspan": 365,
    "min_rb": 0.3,
    "max_fwhm": 5.5,
    "max_elong": 1.4,
    "max_magdiff": 1.0,
    "max_nbad": 2,
    "min_sso_dist": 20,
    "min_gal_lat": 0.,  # Default: 14
    "gaia_rs": 10.,
    "gaia_pm_signif": 3,
    "gaia_plx_signif": 3,
    "gaia_veto_gmag_min": 9,
    "gaia_veto_gmag_max": 20,
    "gaia_excessnoise_sig_max": 999,
    "ps1_sgveto_rad": 1.,
    "ps1_sgveto_th": 0.8,
    "ps1_confusion_rad": 3.,
    "ps1_confusion_sg_tol": 0.1
}

class RetractionError(Exception):
   """Base class for retracted event"""
   pass


class GravWaveScanner(AmpelWizard):

    def __init__(self, gw_name=None, gw_file=None, rev=None, logger=None, prob_threshold=0.95, cone_nside=64,
                 fast_query=False, n_days=None):

        self.prob_threshold = prob_threshold

        if gw_file is None:
            self.gw_path, self.output_path, self.gw_name = self.get_superevent(gw_name, rev)

        else:
            basename = os.path.basename(gw_file)
            self.gw_path = "{0}/{1}".format(base_ligo_dir, basename)
            if gw_file[:8] == "https://":
                logger.info("Downloading from: {0}".format(gw_file))
                self.gw_path = "{0}/{1}".format(base_ligo_dir, os.path.basename(gw_file[7:]))
                wget.download(gw_file, self.gw_path)

            self.output_path = "{0}/{1}_{2}.pdf".format(
                ligo_candidate_output_dir, os.path.basename(gw_file), self.prob_threshold)
            self.gw_name = os.path.basename(gw_file[7:])
        self.data, t_obs, self.hpm, self.key, self.dist, self.dist_unc = self.read_map()

        t_min = Time(t_obs, format="isot", scale="utc")

        logging.info("MERGER TIME: {0}".format(t_min))
        logging.info("Reading map")

        self.pixel_threshold = self.find_pixel_threshold(self.data[self.key])
        self.map_coords, self.pixel_nos, self.map_probs, self.ligo_nside, self.pixel_area = self.unpack_skymap()
        AmpelWizard.__init__(self, run_config=gw_run_config, t_min=t_min, logger=logger, cone_nside=cone_nside,
                             fast_query=fast_query)

        # By default, accept things detected within 72 hours of merger
        if n_days is None:
            n_days = 3.

        self.default_t_max = Time(self.t_min.jd + n_days, format="jd")

    def get_name(self):
        return self.gw_name

    def get_full_name(self):
        return self.gw_name

    # def get_tiling_line(self):
    #
    #     too_bool = len(["ToO" in x for x in self.get_multi_night_summary().data.qcomment])> 0
    #
    #     if too_bool:
    #         return "The tiling was optimally determined and triggered using the GROWTH Target of Opportunity marshal (Coughlin et al. 2019a, Kasliwal et al. 2019b). "
    #     else:
    #         return ""

    def get_obs_line(self):
        return "Each exposure was 30s with a typical depth of 20.5 mag."

    def get_overlap_line(self):
        return "We covered {0:.1f}% of the enclosed probability " \
               "based on the map in {1:.1f} sq deg. " \
               "This estimate accounts for chip gaps. ".format(
            self.overlap_prob, self.area)

    @staticmethod
    def remove_variability_line():
        return ", and removing candidates with history of " \
               "variability prior to the merger time"

    def candidate_text(self, name, first_detection, lul_lim, lul_jd):

        try:
            text = "{0}, first detected {1:.1f} hours after merger, " \
            "was not detected {2:.1f} days prior to a depth of {3:.2f}. ".format(
                name, 24. * (first_detection - self.t_min.jd), first_detection - lul_jd, lul_lim
            )
        except TypeError:
            text = "{0} had upper limit problems. PLEASE FILL IN NUMBERS BY HAND!!! ".format(name)
        return text

    def filter_f_no_prv(self, res):

        # Positive detection
        if res['candidate']['isdiffpos'] not in ["t", "1"]:
            logging.debug("Negative subtraction.")
            return False

        # Veto old transients
        if res["candidate"]["jdstarthist"] < self.t_min.jd:
            logging.debug("Transient is too old. (jdstarthist history predates event)")
            return False

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            logging.debug("Outside of event contour.")
            return False

        logging.debug("Passed filter f (no prv)")

        return True

    # def fast_filter_f_no_prv(self, res):
    #
    #     # Positive detection
    #     if res['candidate']['isdiffpos'] not in ["t", "1"]:
    #         return False
    #
    #     # Veto old transients
    #     if res["candidate"]["jdstarthist"] < self.t_min.jd:
    #         return False
    #
    #     # Require 2 detections separated by 15 mins
    #     if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
    #         return False
    #
    #     return True

    def filter_f_history(self, res):
        # Veto old transients
        if res["candidate"]["jdstarthist"] < self.t_min.jd:
            logging.debug("Transient is too old. (jdstarthist history predates event)")
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > self.default_t_max.jd:
            logging.debug("Transient is too new. (jdstarthist too late after event)")
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            logging.debug("Not passed mover cut")
            return False

        # Require 2 positive detections
        old_detections = [x for x in res["prv_candidates"] if np.logical_and(
            x["isdiffpos"] is not None,
            x["jd"] > self.t_min.jd
        )]

        pos_detections = [x for x in old_detections if x['isdiffpos'] in ["t", "1"]]

        if len(pos_detections) < 1:
            logging.debug("Does not have two detections")
            return False

        return True

    def get_superevent(self, name, rev):
        if name is None:
            superevent_iterator = ligo_client.superevents('category: Production')
            superevent_ids = [superevent['superevent_id'] for superevent in superevent_iterator]
            name = superevent_ids[0]

        voevents = ligo_client.voevents(name).json()["voevents"]

        if rev is None:
            rev = len(voevents)

        elif rev > len(voevents):
            raise Exception("Revision {0} not found".format(rev))

        latest_voevent = voevents[rev - 1]
        print("Found voevent {0}".format(latest_voevent["filename"]))

        if "Retraction" in latest_voevent["filename"]:
            raise RetractionError("The specified LIGO event, {0}, was retracted.".format(latest_voevent["filename"]))

        response = requests.get(latest_voevent["links"]["file"])

        root = lxml.etree.fromstring(response.content)
        params = {elem.attrib['name']:
                      elem.attrib['value']
                  for elem in root.iterfind('.//Param')}

        latest_skymap = params["skymap_fits"]

        print("Latest skymap URL: {0}".format(latest_skymap))

        base_file_name = os.path.basename(latest_skymap)
        savepath = os.path.join(base_ligo_dir, "{0}_{1}_{2}".format(
            name, latest_voevent["N"], base_file_name))

        print("Saving to: {0}".format(savepath))
        response = requests.get(latest_skymap)

        with open(savepath, "wb") as f:
            f.write(response.content)

        output_file = "{0}/{1}_{2}_{3}.pdf".format(ligo_candidate_output_dir, name, latest_voevent["N"],
                                                   self.prob_threshold)

        return savepath, output_file, name

    def read_map(self, ):
        print("Reading file: {0}".format(self.gw_path))
        data, h = fitsio.read(self.gw_path, header=True)#columns=["PROB"],
        if "DISTMEAN" not in h:
            dist = None
        else:
            dist = h["DISTMEAN"]
        if "DISTSTD" not in h:
            dist_unc = None
        else:
            dist_unc = h["DISTSTD"]
        if "DATE-OBS" not in h:
            t_obs = fitsio.read_header(self.gw_path)["DATE-OBS"]
        else:
            t_obs = h["DATE-OBS"]

        if "PROB" in data.dtype.names:
            key = "PROB"
        elif 'PROBABILITY' in data.dtype.names:
            key = 'PROB'
            prob = np.array(data["PROBABILITY"]).flatten()
            data = append_fields(data, "PROB", prob)
        else:
            raise Exception("No recognised probability key in map. This is probably a weird one, right?")

        if not isinstance(data[0], float):
            probs = np.array(data["PROB"]).flatten()
            # # print(data)
            # # print(drop_fields(data, "PROB"))
            # print(data.dtype.names)
            # data = drop_fields(data, "PROB", asrecarray=True)
            # print(type(data))
            data = np.array(probs, dtype=np.dtype([("PROB", np.float)]))

        logging.info(f"Summed probability is {100. * np.sum(data['PROB']):.1f}%")

        if h["ORDERING"] == "RING":
            data["PROB"] = hp.pixelfunc.reorder(data["PROB"], inp="RING", out="NESTED")
            h["ORDERING"] = "NESTED"

        hpm = HEALPix(nside=h["NSIDE"], order=h["ORDERING"], frame='icrs')

        # with fits.open(self.gw_path) as hdul:
        #     print("Opened file")
        #     t_obs = hdul[0].header
        #     print(t_obs)
        #     print("read merger time")
        #     data = hdul[1].data
        #     print("Read data")
        return data, t_obs, hpm, key, dist, dist_unc

    def find_pixel_threshold(self, data):

        ranked_pixels = np.sort(data)[::-1]
        int_sum = 0.0
        pixel_threshold = 0.0

        for i, prob in enumerate(ranked_pixels):
            int_sum += prob
            if int_sum > self.prob_threshold:
                print("Threshold found! \n To reach {0}% of probability, pixels with "
                      "probability greater than {1} are included".format(
                    int_sum * 100., prob))
                pixel_threshold = prob
                break

        return pixel_threshold

    def unpack_skymap(self):

        ligo_nside = hp.npix2nside(len(self.data[self.key]))

        threshold = self.find_pixel_threshold(self.data[self.key])

        mask = self.data[self.key] > threshold

        map_coords = []

        pixel_nos = []

        print("Checking which pixels are within the contour:")

        for i in tqdm(range(hp.nside2npix(ligo_nside))):
            if mask[i]:
                map_coords.append(self.extract_ra_dec(ligo_nside, i))
                pixel_nos.append(i)

        pixel_area = hp.nside2pixarea(ligo_nside, degrees=True) * float(len(map_coords))

        print("Total pixel area: {0} degrees".format(pixel_area))

        map_coords = np.array(map_coords, dtype=np.dtype([("ra", np.float),
                                                          ("dec", np.float)]))

        return map_coords, pixel_nos, self.data[self.key][mask], ligo_nside, pixel_area

    def find_cone_coords(self):
        cone_ids = []

        for ra, dec in self.map_coords:
            cone_ids.append(self.extract_npix(self.cone_nside, ra, dec))

        cone_ids = list(set(cone_ids))

        cone_coords = []

        for i in tqdm(cone_ids):
            cone_coords.append(self.extract_ra_dec(self.cone_nside, i))

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", np.float), ("dec", np.float)])
        )

        return cone_ids, cone_coords

    def plot_skymap(self):
        fig = plt.figure()
        plt.subplot(211, projection="aitoff")

        mask = self.data[self.key] > self.pixel_threshold

        size = hp.max_pixrad(self.ligo_nside, degrees=True) ** 2

        plt.scatter(self.wrap_around_180(self.map_coords["ra"]), self.map_coords["dec"],
                         c=self.data[self.key][mask], vmin=0., vmax=max(self.data[self.key]), s=size)
        plt.title("LIGO SKYMAP")

        plt.subplot(212, projection="aitoff")

        plt.scatter(self.wrap_around_180(self.cone_coords["ra"]), self.cone_coords["dec"])
        plt.title("CONE REGION")
        return fig

    def interpolate_map(self, ra_deg, dec_deg):
        return self.hpm.interpolate_bilinear_skycoord(SkyCoord(ra_deg * u.deg, dec_deg * u.deg), self.data[self.key])

    def in_contour(self, ra_deg, dec_deg):
        return self.interpolate_map(ra_deg, dec_deg) > self.pixel_threshold

if __name__=="__main__":

    import logging
    logger = logging.getLogger("quiet_logger")
    logger.setLevel(logging.INFO)

    gw = GravWaveScanner(logger=logger)
    gw.scan_cones()



import requests
import backoff
from base64 import b64decode
from astropy.time import Time
from nuztf.credentials import load_credentials
from requests.auth import HTTPBasicAuth

# AMPEL API URLs

API_BASEURL = "https://ampel.zeuthen.desy.de"
API_ZTF_ARCHIVE_URL = API_BASEURL + "/api/ztf/archive"
API_CATALOGMATCH_URL = API_BASEURL + "/api/catalogmatch"
API_CUTOUT_URL = API_BASEURL + "/api/ztf/archive/cutouts"

api_user, api_pass = load_credentials("ampel_api")


def merge_alerts(alert_list):
    merged_list = []
    keys = list(set([x["objectId"] for x in alert_list]))

    for objectid in keys:
        alerts = [x for x in alert_list if x["objectId"] == objectid]
        if len(alerts) == 1:
            merged_list.append(alerts[0])
        else:
            jds = [x["candidate"]["jd"] for x in alerts]
            order = [jds.index(x) for x in sorted(jds)[::-1]]
            latest = alerts[jds.index(max(jds))]
            latest["candidate"]["jdstarthist"] = min(
                [x["candidate"]["jdstarthist"] for x in alerts]
            )

            for index in order[1:]:

                x = alerts[index]

                # Merge previous detections

                for prv in x["prv_candidates"] + [x["candidate"]]:
                    if prv not in latest["prv_candidates"]:
                        latest["prv_candidates"] = [prv] + latest["prv_candidates"]

            merged_list.append(latest)
    return merged_list

@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)


def ampel_api_cone(ra, dec, radius,
                   t_min_jd=Time(
                       '2018-04-01T00:00:00.123456789',
                       format='isot',
                       scale='utc'
                   ).jd,
                   t_max_jd=Time.now().jd, logger=None):
    """Function to query ampel via a cone search"""

    queryurl_conesearch = (
            API_ZTF_ARCHIVE_URL
            + f"/alerts/cone_search?ra={ra}&dec={dec}&"
              f"radius={radius}&jd_start={t_min_jd}&"
              f"jd_end={t_max_jd}&with_history=false&"
              f"with_cutouts=false&chunk_size=500"
    )

    if logger is not None:
        logger.debug(queryurl_conesearch)

    response = requests.get(
        queryurl_conesearch,
        auth=HTTPBasicAuth(api_user, api_pass),
    )
    if response.status_code == 503:
        raise requests.exceptions.RequestException

    query_res = [i for i in response.json()["alerts"]]

    return query_res

@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)


def ampel_api_name(ztf_name, logger=None):
    """Function to query ampel via name"""

    queryurl_ztf_name = (
        API_ZTF_ARCHIVE_URL + f"/object/{ztf_name}/alerts?with_history=true"
    )

    if logger is not None:
        logger.debug(queryurl_ztf_name)

    response = requests.get(
        queryurl_ztf_name,
        auth=HTTPBasicAuth(api_user, api_pass),
    )
    if response.status_code == 503:
        raise requests.exceptions.RequestException
    query_res = [i for i in response.json()]
    query_res = merge_alerts(query_res)
    return query_res


@backoff.on_exception(
    backoff.expo,
    requests.exceptions.RequestException,
    max_time=600,
)

def ampel_api_cutout(candid: int, logger=None):
    """Function to query ampel for cutouts by candidate ID"""
    queryurl_cutouts = API_CUTOUT_URL + f"/{candid}"
    response = requests.get(
        queryurl_cutouts,
        auth=HTTPBasicAuth(api_user, api_pass),
    )
    if logger is not None:
        self.logger.debug(queryurl_cutouts)

    if response.status_code == 503:
        raise requests.exceptions.RequestException

    cutouts = response.json()
    return cutouts


def reassemble_alert(mock_alert):
    """Function to recreate ztf alerts"""
    cutouts = ampel_api_cutout(mock_alert["candid"])

    for k in cutouts:
        mock_alert[f"cutout{k.title()}"] = {
            "stampData": b64decode(cutouts[k]),
            "fileName": "dunno",
        }

    mock_alert["schemavsn"] = "dunno"
    mock_alert["publisher"] = "dunno"
    for pp in [mock_alert["candidate"]] + mock_alert["prv_candidates"]:
        pp["pdiffimfilename"] = "dunno"
        pp["programpi"] = "dunno"
        pp["ssnamenr"] = "dunno"

    return mock_alert


def ampel_api_tns(ra: float, dec: float, searchradius_arcsec: float = 3):
    """Function to query TNS via ampel api"""
    queryurl_catalogmatch = API_CATALOGMATCH_URL + f"/cone_search/nearest"

    # First, we create a json body to post
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    query = {
        "ra_deg": ra,
        "dec_deg": dec,
        "catalogs": [
            {"name": "TNS", "rs_arcsec": searchradius_arcsec, "use": "extcats"}
        ],
    }

    # Now we retrieve results from the API
    response = requests.post(url=queryurl_catalogmatch, json=query, headers=headers)

    full_name = None
    discovery_date = None
    source_group = None

    if response.json()[0]:
        response_body = response.json()[0]["body"]
        print(response_body)
        name = response_body["objname"]
        prefix = response_body["name_prefix"]
        full_name = prefix + name
        discovery_date = response_body["discoverydate"]
        if "source_group" in response_body.keys():
            source_group = response_body["source_group"]["group_name"]

    return full_name, discovery_date, source_group


def ampel_api_catalog(
        catalog: str,
        catalog_type: str,
        ra: float,
        dec: float,
        searchradius_arcsec: float = 10,
        searchtype: str = "all",
    ):
    """
    Method for querying catalogs via the Ampel API
    'catalog' must be the name of a supported catalog, e.g.
    SDSS_spec, PS1, NEDz_extcats...
    For a full list of catalogs, confer
    https://ampel.zeuthen.desy.de/api/catalogmatch/catalogs

    """
    assert catalog_type in ["extcats", "catsHTM"]
    assert searchtype in ["all", "nearest"]

    queryurl_catalogmatch = API_CATALOGMATCH_URL + "/cone_search/" + searchtype

    # First, we create a json body to post
    headers = {"accept": "application/json", "Content-Type": "application/json"}
    query = {
        "ra_deg": ra,
        "dec_deg": dec,
        "catalogs": [
            {"name": catalog, "rs_arcsec": searchradius_arcsec, "use": catalog_type}
        ],
    }

    response = requests.post(
        url=queryurl_catalogmatch, json=query, headers=headers
    ).json()[0]

    return response



#!/usr/bin/env python3
# coding: utf-8

from nuztf.ampel_magic import AmpelWizard
from astropy.time import Time
import healpy as hp
import numpy as np
from tqdm import tqdm
import os
from pathlib import Path
import requests
import logging
from ztf_plan_obs import gcn_parser

nu_candidate_output_dir = os.path.join(Path(__file__).resolve().parents[1], "Neutrino_candidates")

nu_run_config = {
    "min_ndet": 1,  # Default:2
    "min_tspan": -1,  # Default 0, but that rejects everything!
    "max_tspan": 365,
    "min_rb": 0.0,
    "max_fwhm": 5.5,
    "max_elong": 1.4,
    "max_magdiff": 1.0,
    "max_nbad": 2,
    "min_sso_dist": 20,
    "min_gal_lat": -1.0,  # Default: 14
    "gaia_rs": 20,
    "gaia_pm_signif": 3,
    "gaia_plx_signif": 3,
    "gaia_veto_gmag_min": 9,
    "gaia_veto_gmag_max": 20,
    "gaia_excessnoise_sig_max": 999,
    "ps1_sgveto_rad": 1,
    "ps1_sgveto_th": 0.8,
    "ps1_confusion_rad": 3,
    "ps1_confusion_sg_tol": 0.1,
}


class ParsingError(Exception):
    """Base class for parsing error"""

    pass


class NeutrinoScanner(AmpelWizard):
    def __init__(
        self,
        nu_name=None,
        manual_args=None,
        gcn_no=None,
        logger=None,
        cone_nside=128,
        t_precursor=None,
    ):

        self.prob_threshold = 0.9

        if manual_args is None:

            if nu_name is not None:
                gcn_no = self.find_gcn_no(nu_name)

            elif gcn_no is None:
                gcn_no = self.get_latest_gcn()

            # nu_name, author, ra, dec, nu_time = self.parse_gcn(gcn_no)

            gcn_info = gcn_parser.parse_gcn_circular(gcn_no)
            print(gcn_info)

            nu_name = gcn_info["name"]
            author = gcn_info["author"]
            ra = [gcn_info["ra"], gcn_info["ra_err"][0], gcn_info["ra_err"][1]]
            dec = [gcn_info["dec"], gcn_info["dec_err"][0], gcn_info["dec_err"][1]]
            nu_time = gcn_info["time"]

            if t_precursor is not None:
                nu_time = Time(nu_time.mjd - t_precursor, format="mjd")

        else:
            (nu_name, ra, dec, nu_time) = manual_args
            author = None

        self.nu_name = nu_name
        self.author = author
        self.gcn_no = gcn_no
        self.dist = None

        print(f"Neutrino time: {nu_time}")

        self.ra_max = float(max(ra[1:]) + ra[0])
        self.ra_min = float(min(ra[1:]) + ra[0])
        self.dec_max = float(max(dec[1:]) + dec[0])
        self.dec_min = float(min(dec[1:]) + dec[0])

        print(f"Coordinates: RA = {ra[0]} ({self.ra_min} - {self.ra_max})")
        print(f"Coordinates: DEC = {dec[0]} ({self.dec_min} - {self.dec_max})")

        self.output_path = f"{nu_candidate_output_dir}/{nu_name}.pdf"
        AmpelWizard.__init__(
            self,
            t_min=nu_time,
            run_config=nu_run_config,
            logger=logger,
            cone_nside=cone_nside,
        )
        self.prob_threshold = 0.9
        self.area = (
            (self.ra_max - self.ra_min)
            * (self.dec_max - self.dec_min)
            * abs(np.cos(np.radians(dec[0])))
        )
        print(f"Projected Area: {self.area}")
        (
            self.map_coords,
            self.pixel_nos,
            self.nside,
            self.map_probs,
            self.data,
            self.key,
        ) = self.unpack_map()

    @staticmethod
    def gcn_url(gcn_number):
        return f"https://gcn.gsfc.nasa.gov/gcn3/{gcn_number}.gcn3"

    def get_name(self):
        return self.nu_name

    def get_full_name(self):
        return f"neutrino event {self.get_name()} ({self.author} et. al, GCN {self.gcn_no})"

    def get_overlap_line(self):
        return (
            f"We covered {self.area:.1f} sq deg, corresponding to {self.overlap_prob:.1f}% of the reported localization region. "
            "This estimate accounts for chip gaps. "
        )

    def candidate_text(self, name, first_detection, lul_lim, lul_jd):
        fd = Time(first_detection, format="mjd")

        text = f"{name} was first detected on {fd.utc}. "

        return text

    # @staticmethod
    # def get_tiling_line():
    #     return ""

    @staticmethod
    def get_obs_line():
        return "Each exposure was 300s with a typical depth of 21.0 mag."

    @staticmethod
    def remove_variability_line():
        return ""

    @staticmethod
    def strip_numbers(line):
        vals = []
        line = line.replace("- ", "-")
        line = line.replace("-", " -")

        for x in line.replace(",", " ").replace("/", " ").split(" "):
            try:
                vals.append(
                    float(
                        "".join(
                            [y for y in x if y not in ["(", "+", ")", "[", "]", "/"]]
                        )
                    )
                )
            except ValueError:
                pass
        return vals

    # def parse_gcn(self, gcn_number):
    #     url = self.gcn_url(gcn_number)
    #     page = requests.get(url)
    #     print(f"Found GCN: {url}")
    #     name = author = ra = dec = time = None
    #     print(page.text)
    #     for line in page.text.splitlines():
    #         line = "".join([x for x in line if x not in ["Â"]])
    #         if "SUBJECT" in line:
    #             name = line.split(" - ")[0].split(": ")[1]
    #         elif "FROM" in line:
    #             base = line.split("at")[0].split(": ")[1].split(" ")
    #             author = [x for x in base if x != ""][1]
    #         elif np.logical_and(np.sum([x in line for x in ["Ra", "RA"]]) > 0, ra is None):
    #             ra = self.strip_numbers(line)
    #         elif np.logical_and(np.sum([x in line for x in ["Dec", "DEC"]]) > 0, dec is None):
    #             dec = self.strip_numbers(line)
    #         elif np.logical_and(np.sum([x in line for x in ["Time", "TIME"]]) > 0, dec is None):
    #             raw_time = [x for x in  line.split(" ") if x not in ["Time", "", "UT", "UTC"]][1]
    #             raw_time = "".join([x for x in raw_time if np.logical_or(
    #                 x.isdigit(), x in [":", "."]
    #             )])
    #             raw_date = name.split("-")[1][:6]
    #             ut_time = f"20{raw_date[0:2]}-{raw_date[2:4]}-{raw_date[4:6]}T{raw_time}"
    #             time = Time(ut_time, format='isot', scale='utc')

    #     try:

    #         if np.sum([x is not None for x in [name, author, ra, dec, time]]) == 5:
    #             ra2 = ra[1]
    #             ra2 = ra2*-1
    #             ra.append(ra2)
    #             if np.logical_and(len(ra) == 3, len(dec) == 3):
    #                 print(name, author, ra, dec, time)
    #                 return name, author, ra, dec, time

    #     except:
    #         pass

    #     raise ParsingError(f"Error parsing GCN {0}".format(url))

    def parse_gcn_archive(self):
        page = requests.get("https://gcn.gsfc.nasa.gov/gcn3_archive.html")

        nu_circulars = []

        for line in page.text.splitlines():
            if "IceCube observation of a high-energy neutrino" in line:
                res = line.split(">")
                gcn_no = "".join([x for x in res[2] if x.isdigit()])
                name = res[3].split(" - ")[0]
                nu_circulars.append((name, gcn_no))

        return nu_circulars

    @staticmethod
    def parse_gcn_for_no(
        base_nu_name, url="https://gcn.gsfc.nasa.gov/gcn3_archive.html"
    ):

        print(f"Checking for GCN on {url}")

        nu_name = str(base_nu_name)

        page = requests.get(url)

        gcn_no = None
        name = None

        while not nu_name[0].isdigit():
            nu_name = nu_name[1:]

        latest_archive_no = None

        for line in page.text.splitlines():
            if np.logical_and(
                "IceCube observation of a high-energy neutrino" in line, nu_name in line
            ):
                res = line.split(">")
                if gcn_no is None:
                    gcn_no = "".join([x for x in res[2] if x.isdigit()])
                    name = res[3].split(" - ")[0]
                    print(f"Found match to {base_nu_name}: {name}")
                else:
                    raise Exception(f"Multiple matches found to {base_nu_name}")

            elif np.logical_and("gcn3_arch_old" in line, latest_archive_no is None):
                url = line.split('"')[1]
                latest_archive_no = int(url[13:].split(".")[0])

        return gcn_no, name, latest_archive_no

    def find_gcn_no(self, base_nu_name):

        gcn_no, name, latest_archive_no = self.parse_gcn_for_no(base_nu_name)

        if gcn_no is None:

            print(
                f"No GCN found for {base_nu_name} on GCN page, checking archive instead. "
                f"The latest page is {latest_archive_no}"
            )

            while np.logical_and(latest_archive_no > 0, gcn_no is None):
                gcn_no, name, _ = self.parse_gcn_for_no(
                    base_nu_name,
                    url=f"https://gcn.gsfc.nasa.gov/gcn3_arch_old{latest_archive_no}.html",
                )
                latest_archive_no -= 1

        # while

        if name is None:
            raise ParsingError("No GCN match found for {0}".format(base_nu_name))

        print(f"Match is {name} (GCN #{gcn_no})")

        return gcn_no

    def get_latest_gcn(self):
        latest = self.parse_gcn_archive()[0]
        print(f"Latest GCN is {latest[0]} (GCN #{latest[1]})")
        return latest[1]

    def filter_f_no_prv(self, res):

        # Positive detection
        if res["candidate"]["isdiffpos"] not in ["t", "1"]:
            logging.debug("Negative subtraction")
            return False

        try:
            if res["candidate"]["drb"] < 0.3:
                logging.debug("DRB too low")
                return False
        except (KeyError, TypeError) as e:
            pass

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            logging.debug("Not in contour")
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            logging.debug("Does not have 2 detections separated  by >15 mins")
            return False

        return True

    def filter_f_history(self, res):
        # Require 2 detections

        n_detections = len(
            [x for x in res["prv_candidates"] if "isdiffpos" in x.keys()]
        )

        if n_detections < 1:
            logging.debug("{0} has insufficient detection".format(res["objectId"]))
            return False

        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            logging.debug("{0} not in contour".format(res["objectId"]))
            return False

        return True

    def find_cone_coords(self):

        cone_coords = []
        cone_ids = []

        scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))

        print("Finding search pixels:")

        for i in tqdm(range(hp.nside2npix(self.cone_nside))):
            ra, dec = self.extract_ra_dec(self.cone_nside, i)
            ra_deg = np.degrees(ra)
            dec_deg = np.degrees(dec)
            if np.logical_and(
                ra_deg > self.ra_min - scan_radius, ra_deg < self.ra_max + scan_radius
            ):
                if np.logical_and(
                    dec_deg > self.dec_min - scan_radius,
                    dec_deg < self.dec_max + scan_radius,
                ):
                    cone_coords.append((ra, dec))
                    cone_ids.append(i)

        cone_coords = np.array(
            cone_coords, dtype=np.dtype([("ra", np.float), ("dec", np.float)])
        )

        return cone_ids, cone_coords

    def in_contour(self, ra_deg, dec_deg):

        in_ra = np.logical_and(ra_deg > self.ra_min, ra_deg < self.ra_max)
        in_dec = np.logical_and(dec_deg > self.dec_min, dec_deg < self.dec_max)

        return np.logical_and(in_ra, in_dec)

    def unpack_map(self):

        # nside = self.cone_nside
        nside = 1024
        map_coords = []
        pixel_nos = []

        center_ra = np.radians(np.mean([self.ra_max, self.ra_min]))
        center_dec = np.radians(np.mean([self.dec_max, self.dec_min]))
        rad = (
            np.radians(max(self.ra_max - self.ra_min, self.dec_max - self.dec_min))
            / 2.0
        )

        nearish_pixels = list(
            hp.query_disc(
                nside=nside,
                vec=hp.ang2vec(np.pi / 2.0 - center_dec, center_ra),
                radius=rad,
                nest=True,
            )
        )

        for i in tqdm(nearish_pixels):
            ra, dec = self.extract_ra_dec(nside, i)
            if self.in_contour(np.degrees(ra), np.degrees(dec)):
                map_coords.append((ra, dec))
                pixel_nos.append(i)

        map_probs = np.ones_like(pixel_nos, dtype=np.float)
        map_probs /= np.sum(map_probs)

        key = "PROB"

        data = np.zeros(hp.nside2npix(nside), dtype=np.dtype([(key, np.float)]))
        data[np.array(pixel_nos)] = map_probs

        return map_coords, pixel_nos, nside, map_probs, data, key


if __name__ == "__main__":

    import logging

    logger = logging.getLogger("quiet_logger")
    logger.setLevel(logging.ERROR)

    nu = NeutrinoScanner(logger=logger)
    nu.scan_cones()



import os
import logging
from ztfquery import io

# Manage ztfquery logins from environment variables


def load_credentials(name):
    """ZTFquery wrapper for loading credentials.
    """
    return io._load_id_(name)


try:
    io.set_account("ampel_api",
                   username=os.environ["AMPEL_API_USER"],
                   password=os.environ["AMPEL_API_PASSWORD"])
    logging.info('Set up "ampel_api" credentials')

except KeyError:
    logging.info('No Credentials for AMPEL API found in environment'
                 'Assume they are set.')

try:
    io.set_account("irsa",
                   username=os.environ["IRSA_USER"],
                   password=os.environ["IRSA_PASSWORD"])
    logging.info('Set up "irsa" credentials')

except KeyError:
    logging.info('No Credentials for "irsa" found in environment'
                 'Assuming they are set.')

try:
    io.set_account("skyvision",
                   username=os.environ["SKYVISION_USER"],
                   password=os.environ["SKYVISION_PASSWORD"])
    logging.info('Set up "skyvision" credentials')

except KeyError:
    logging.info('No Credentials for "skyvision" found in environment'
                 'Assuming they are set.')



import pickle
import argparse
from multiprocessing import JoinableQueue, Process
from nuztf.gw_scanner import GravWaveScanner
import numpy as np
import healpy as hp
import os
from tqdm import tqdm
from astropy.time import Time
from pathlib import Path

ligo_candidate_cache = os.path.join(Path(__file__).resolve().parents[1], "LIGO_cache")

class MultiGwProcessor(GravWaveScanner):
    queue = None
    results = dict()

    def __init__(self, n_cpu=os.cpu_count()-1, mp_id=0, n_days=None, *args, **kwargs):

        print("Running on {0} CPUs".format(n_cpu))

        GravWaveScanner.__init__(self, n_days=n_days, *args, **kwargs)
        self.fill_queue, self.n_sky, self.scan_method = self.optimise_scan_method()

        self.cache_dir = os.path.join(
            ligo_candidate_cache,
            os.path.splitext(os.path.basename(self.output_path))[0]
        )
        try:
            os.makedirs(self.cache_dir)
        except OSError:
            pass
        self.scan_radius = np.degrees(hp.max_pixrad(self.cone_nside))
        self.queue = JoinableQueue()
        self.processes = [Process(target=self.scan_wrapper, kwargs={"mp_id": i+1}) for i in range(int(n_cpu))]

        self.obj_names = []
        self.mp_id = mp_id

        for p in self.processes:
            p.start()

    def add_to_queue(self, item):
        self.queue.put(item)

    def scan_wrapper(self, **kwargs):
        self.mp_id = kwargs["mp_id"]

        while True:
            item = self.queue.get()
            if item is None:
                break

            (j, mts, query_res) = item

            # print("{0} of {1} queries: Staring {2} alerts".format(j, mts, len(query_res)))

            res = self.filter(query_res)

            # print("{0} of {1} queries: {2} accepted out of {3} alerts".format(j, mts, len(res), len(query_res)))

            self.obj_names += [x["objectId"] for x in res]
            if len(res) > 0:
                self.dump_cache()

            # self.dump_cache(res)
            self.queue.task_done()

    def filter(self, query_res):

        indexes = []

        for i, res in enumerate(query_res):
            if self.filter_f_no_prv(res):
                indexes.append(i)

        return [query_res[i] for i in indexes]

    def filter_f_no_prv(self, res):

        # Veto old transients
        if res["candidate"]["jdstarthist"] < self.t_min.jd:
            logging.debug("Transient is too old")
            return False

        # Veto new transients
        if res["candidate"]["jdstarthist"] > self.default_t_max.jd:
            logging.debug("Transient is too new")
            return False

        # Positive detection
        if res['candidate']['isdiffpos'] not in ["t", "1"]:
            logging.debug("Negative subtraction")
            return False

        try:
            if res['candidate']['drb'] < 0.3:
                logging.debug("DRB too low")
                return False
        except KeyError:
            pass
        except TypeError:
            pass

        # Check contour
        if not self.in_contour(res["candidate"]["ra"], res["candidate"]["dec"]):
            logging.debug("Not in contour")
            return False

        # Require 2 detections separated by 15 mins
        if (res["candidate"]["jdendhist"] - res["candidate"]["jdstarthist"]) < 0.01:
            logging.debug("Does not have 2 detections separated  by >15 mins")
            return False

        return True

    def dump_cache(self):

        path = os.path.join(self.cache_dir, "{0}.pkl".format(self.mp_id))

        with open(path, "wb") as f:
            pickle.dump(self.obj_names, f)

    def optimise_scan_method(self, t_max=None):

        if t_max is None:
            t_max = Time.now()

        length_window = t_max.jd - self.t_min.jd

        survey_start_jd = 2458000.

        full_ztf = t_max.jd - survey_start_jd

        n_skys_time = length_window * 1./4.

        n_skys_space = full_ztf * self.pixel_area/1000.

        if n_skys_time < n_skys_space:
            f = self.fill_queue_time
            method = "time"
        else:
            f = self.fill_queue_space
            method  = "space"

        n_sky = min([n_skys_space, n_skys_time])

        print("Scanning method: {0} \n N_sky: {1}".format(method, n_sky))

        return f, n_sky, method

    def fill_queue_time(self, t_max=None):

        if t_max is None:
            t_max = Time(self.default_t_max.jd + 7., format="jd")

        time_steps = np.arange(self.t_min.jd, t_max.jd, step=0.005)
        mts = len(time_steps)

        n_tot = 0

        print("Scanning between {0}JD and {1}JD".format(time_steps[0], time_steps[-1]))

        for j, t_start in enumerate(tqdm(list(time_steps[:-1]))):

            ztf_object = ampel_client.get_alerts_in_time_range(
                jd_min=t_start, jd_max=time_steps[j+1], with_history=False)
            query_res = [x for x in ztf_object]
            n_tot += len(query_res)
            self.add_to_queue((j, mts, query_res))
            self.scanned_pixels.append(j)

        print("Added {0} candidates since {1}".format(n_tot, time_steps[0]))

    def fill_queue_space(self, t_max=None):

        if t_max is None:
            t_max = Time.now()

        mts = len(list(self.cone_ids))
        n_tot = 0

        for j, cone_id in enumerate(tqdm(list(self.cone_ids))):
            ra, dec = self.cone_coords[j]

            if cone_id not in self.scanned_pixels:
                ztf_object = ampel_client.get_alerts_in_cone(
                    ra, dec, self.scan_radius, self.t_min.jd, t_max.jd, with_history=False)
                query_res = [x for x in ztf_object]
                n_tot += len(query_res)
                self.add_to_queue((j, mts, query_res))
                self.scanned_pixels.append(j)

        print("Added {0} candidates since {1}".format(n_tot, self.t_min.jd))

    def terminate(self):
        """ wait until queue is empty and terminate processes """
        self.queue.join()
        for p in self.processes:
            p.terminate()

    def combine_cache(self):

        for name in self.get_cache_file():
            try:
                with open(os.path.join(self.cache_dir, name), "rb") as f:
                    self.obj_names += pickle.load(f)
            except:
                pass

        self.obj_names = list(set(self.obj_names))

        print("Scanned {0} pixels".format(len(self.scanned_pixels)))
        print("Found {0} candidates passing the first filtering stage.".format(len(self.obj_names)))
        print(self.obj_names)

        ztf_object = ampel_client.get_alerts_for_object(self.obj_names, with_history=True)

        query_res = [i for i in ztf_object]

        query_res = self.merge_alerts(query_res)

        self.cache = dict()

        for res in tqdm(query_res):
            if self.filter_f_history(res):
                if self.filter_ampel(res):
                    self.cache[res["objectId"]] = res
                else:
                    print("Failed Ampel")
            else:
                print("Failed History")

        print("Found {0} candidates passing the final filtering stage.".format(len(self.cache)))

        self.create_candidate_summary()

    def get_cache_file(self):
        return [os.path.join(self.cache_dir, x) for x in os.listdir(self.cache_dir) if ".pkl" in x]

    def clean_cache(self):

        for name in self.get_cache_file():
            os.remove(name)

        print("Cache cleaned!")

if __name__ == '__main__':
    import os
    import logging
    from ampel.pipeline.logging.ExtraLogFormatter import ExtraLogFormatter

    logging.basicConfig()
    logging.root.handlers[0].setFormatter(ExtraLogFormatter())
    logger = logging.getLogger()
    logger.setLevel(logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument("--n_cpu", default=min(24, max(1, os.cpu_count()-1)))
    parser.add_argument("-p", "--prob_threshold", default=0.9, type=float)
    parser.add_argument("-n", "--name", default=None)
    cfg = parser.parse_args()

    print("N CPU available", os.cpu_count())
    print("Using {0} CPUs".format(cfg.n_cpu))

    gw = MultiGwProcessor(gw_name=cfg.name, logger=logger, prob_threshold=cfg.prob_threshold,
                          n_cpu=cfg.n_cpu, fast_query=True)
    gw.clean_cache()
    gw.fill_queue()
    gw.terminate()
    gw.combine_cache()
    gw.clean_cache()
    # gw.plot_overlap_with_observations()